/* 
 * link.c --
 *
 *	This file implements linked variables (a C variable that is
 *	tied to a Tcl variable).  The idea of linked variables was
 *	first suggested by Andreas Stolcke and this implementation is
 *	based heavily on a prototype implementation provided by
 *	him.
 *
 *	Modified by DKF to handle linking to/from Tcl_Objs
 *
 * Copyright (c) 1993 The Regents of the University of California.
 * Copyright (c) 1994-1996 Sun Microsystems, Inc.
 * Copyright (c) 1997 Donal K. Fellows
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: $Id: link.c,v 1.1 1997/05/18 20:55:36 donal Exp fellowsd $
 */

#include "link.h"

/*
 * For each linked variable there is a data structure of the following
 * type, which describes the link and is the clientData for the trace
 * set on the Tcl variable.
 */

typedef struct objLink {
    Tcl_Interp *interp;		/* Interpreter containing Tcl variable. */
    Tcl_Obj *varName;		/* Name of variable (must be global).  This
				 * is needed during trace callbacks, since
				 * the actual variable may be aliased at
				 * that time via upvar. */
    Tcl_Obj **addr;		/* Location of C variable. */
    int flags;			/* Miscellaneous one-bit values;  see below
				 * for definitions. */
} objLink;

/*
 * Definitions for flag bits:
 * LINK_READ_ONLY -		1 means errors should be generated if Tcl
 *				script attempts to write variable.
 * LINK_BEING_UPDATED -		1 means that a call to Dkfobj_UpdateLinkedVar
 *				is in progress for this variable, so
 *				trace callbacks on the variable should
 *				be ignored.
 */

#define LINK_READ_ONLY		1
#define LINK_BEING_UPDATED	2

/*
 * Forward references to procedures defined later in this file:
 */

static char *		ObjLinkTraceProc _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, char *name1, char *name2,
			    int flags));

/*
 *----------------------------------------------------------------------
 *
 * Dkfobj_LinkVar --
 *
 *	Link a C variable to a Tcl variable so that changes to either
 *	one causes the other to change.
 *
 * Results:
 *	The return value is TCL_OK if everything went well or TCL_ERROR
 *	if an error occurred (interp->result is also set after errors).
 *
 * Side effects:
 *	The value at *addr is linked to the Tcl variable "varName".
 *
 *----------------------------------------------------------------------
 */

int
Dkfobj_LinkVar(interp, varName, addr, read_only)
     Tcl_Interp *interp;	/* Interpreter in which varName exists. */
     Tcl_Obj *varName;		/* Name of a global variable in interp. */
     Tcl_Obj **addr;		/* Address of a C variable to be linked
				 * to varName. */
     int read_only;		/* Is the C variable read-only */
{
    objLink *linkPtr;
    int code;

    linkPtr = (objLink *) Tcl_Alloc(sizeof(objLink));
    linkPtr->interp = interp;
    Tcl_IncrRefCount(linkPtr->varName = varName);
    linkPtr->addr = addr;
    linkPtr->flags = read_only?LINK_READ_ONLY:0;

    if (Tcl_ObjSetVar2(interp, varName, NULL, *addr,
		       TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG
		       |TCL_PARSE_PART1) == NULL) {
      Tcl_DecrRefCount(linkPtr->varName);
      Tcl_Free((char *) linkPtr);
      return TCL_ERROR;
    }
    code = Tcl_TraceVar(interp, Tcl_GetStringFromObj(varName, NULL),
			TCL_GLOBAL_ONLY|TCL_TRACE_READS|TCL_TRACE_WRITES
			|TCL_TRACE_UNSETS, ObjLinkTraceProc,
			(ClientData)linkPtr);
    if (code != TCL_OK) {
      Tcl_DecrRefCount(linkPtr->varName);
      Tcl_Free((char *) linkPtr);
    }
    return code;
}

/*
 *----------------------------------------------------------------------
 *
 * Dkfobj_UnlinkVar --
 *
 *	Destroy the link between a Tcl variable and a C variable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	If "varName" was previously linked to a C variable, the link
 *	is broken to make the variable independent.  If there was no
 *	previous link for "varName" then nothing happens.
 *
 *----------------------------------------------------------------------
 */

void
Dkfobj_UnlinkVar(interp, varName)
     Tcl_Interp *interp;	/* Interpreter containing variable to unlink */
     Tcl_Obj *varName;		/* Global variable in interp to unlink. */
{
  objLink *linkPtr;
  char *vName = Tcl_GetStringFromObj(varName, NULL);

  linkPtr = (objLink *) Tcl_VarTraceInfo(interp, vName, TCL_GLOBAL_ONLY,
					 ObjLinkTraceProc, (ClientData) NULL);
  if (linkPtr == NULL) {
    return;
  }
  Tcl_UntraceVar(interp, vName,
	    TCL_GLOBAL_ONLY|TCL_TRACE_READS|TCL_TRACE_WRITES|TCL_TRACE_UNSETS,
	    ObjLinkTraceProc, (ClientData) linkPtr);
  Tcl_DecrRefCount(linkPtr->varName);
  Tcl_Free((char *) linkPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Dkfobj_UpdateLinkedVar --
 *
 *	This procedure is invoked after a linked variable has been
 *	changed by C code.  It updates the Tcl variable so that
 *	traces on the variable will trigger.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The Tcl variable "varName" is updated from its C value,
 *	causing traces on the variable to trigger.
 *
 *----------------------------------------------------------------------
 */

void
Dkfobj_UpdateLinkedVar(interp, varName)
     Tcl_Interp *interp;	/* Interpreter containing variable. */
     Tcl_Obj *varName;		/* Name of global variable that is linked. */
{
  objLink *linkPtr;
  int savedFlag;

  linkPtr = (objLink *)
    Tcl_VarTraceInfo(interp, Tcl_GetStringFromObj(varName, NULL),
		     TCL_GLOBAL_ONLY, ObjLinkTraceProc, (ClientData) NULL);
  if (linkPtr == NULL) {
    return;
  }
  savedFlag = linkPtr->flags & LINK_BEING_UPDATED;
  linkPtr->flags |= LINK_BEING_UPDATED;
  Tcl_ObjSetVar2(interp, linkPtr->varName, NULL, *linkPtr->addr,
		 TCL_GLOBAL_ONLY|TCL_PARSE_PART1);
  linkPtr->flags = (linkPtr->flags & ~LINK_BEING_UPDATED) | savedFlag;
}

/*
 *----------------------------------------------------------------------
 *
 * ObjLinkTraceProc --
 *
 *	This procedure is invoked when a linked Tcl variable is read,
 *	written, or unset from Tcl.  It's responsible for keeping the
 *	C variable in sync with the Tcl variable.
 *
 * Results:
 *	If all goes well, NULL is returned; otherwise an error message
 *	is returned.
 *
 * Side effects:
 *	The C variable may be updated to make it consistent with the
 *	Tcl variable, or the Tcl variable may be overwritten to reject
 *	a modification.
 *
 *----------------------------------------------------------------------
 */

static char *
ObjLinkTraceProc(clientData, interp, name1, name2, flags)
     ClientData clientData;	/* Contains information about the link. */
     Tcl_Interp *interp;	/* Interpreter containing Tcl variable. */
     char *name1;		/* First part of variable name. */
     char *name2;		/* Second part of variable name. */
     int flags;			/* Miscellaneous additional information. */
{
  objLink *linkPtr = (objLink *) clientData;
  Tcl_Obj *value;

  /*
   * If the variable is being unset, then just re-create it (with a
   * trace) unless the whole interpreter is going away.
   */

  if (flags & TCL_TRACE_UNSETS) {
    if (flags & TCL_INTERP_DESTROYED) {
      Tcl_DecrRefCount(linkPtr->varName);
      Tcl_Free((char *) linkPtr);
    } else if (flags & TCL_TRACE_DESTROYED) {
      char *vName = Tcl_GetStringFromObj(linkPtr->varName, NULL);

      Tcl_ObjSetVar2(interp, linkPtr->varName, NULL, *linkPtr->addr,
		     TCL_GLOBAL_ONLY|TCL_PARSE_PART1);
      Tcl_TraceVar(interp, vName, TCL_GLOBAL_ONLY|TCL_TRACE_READS
		   |TCL_TRACE_WRITES|TCL_TRACE_UNSETS,
		   ObjLinkTraceProc, (ClientData) linkPtr);
    }
    return NULL;
  }

    /*
     * If we were invoked because of a call to Dkfobj_UpdateLinkedVar, then
     * don't do anything at all.  In particular, we don't want to get
     * upset that the variable is being modified, even if it is
     * supposed to be read-only.
     */

  if (linkPtr->flags & LINK_BEING_UPDATED) {
    return NULL;
  }

  /*
   * For read accesses, update the Tcl variable if the C variable
   * has changed since the last time we updated the Tcl variable.
   */

  if (flags & TCL_TRACE_READS) {
    Tcl_ObjSetVar2(interp, linkPtr->varName, NULL, *linkPtr->addr,
		   TCL_GLOBAL_ONLY|TCL_PARSE_PART1);
    return NULL;
  }

  /*
   * For writes, first make sure that the variable is writable.  Then
   * convert the Tcl value to C if possible.  If the variable isn't
   * writable or can't be converted, then restore the varaible's old
   * value and return an error.  Another tricky thing: we have to save
   * and restore the interpreter's result, since the variable access
   * could occur when the result has been partially set.
   */

  if (linkPtr->flags & LINK_READ_ONLY) {
    Tcl_ObjSetVar2(interp, linkPtr->varName, NULL, *linkPtr->addr,
		   TCL_GLOBAL_ONLY|TCL_PARSE_PART1);
    return "linked variable is read-only";
  }
  value = Tcl_ObjGetVar2(interp, linkPtr->varName, NULL,
			 TCL_GLOBAL_ONLY|TCL_PARSE_PART1);
  if (value == NULL) {
    /*
     * This shouldn't ever happen.
     */
    return "internal error: linked variable couldn't be read";
  }

  /* Make sure that we increment the refcount first so that if we are
   * updating linkPtr->addr to be itself, we don't accidentally delete
   * the contents in the process, which would be embarrasing... */
  Tcl_IncrRefCount(value);
  Tcl_DecrRefCount(*linkPtr->addr);
  *linkPtr->addr = value;
  return NULL;
}
